/*
 * Config file.
 * Must be include before any other header.
 *
 * Copyright IBM Corp. 2022
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 *
 */
#ifndef LIBPV_CONFIG_H
#define LIBPV_CONFIG_H
#define GETTEXT_PACKAGE "libpv"

#endif /* LIBPV_CONFIG_H */
