/*
 * zdev - Modify and display the persistent configuration of devices
 *
 * Copyright IBM Corp. 2016, 2017
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef GENERIC_CCW_H
#define GENERIC_CCW_H

struct devtype;

extern struct devtype generic_ccw_devtype;

#endif /* GENERIC_CCW_H */
