/*
 * SPDX-License-Identifier: MIT
 *
 * Copyright IBM Corp.
 */

#ifndef LIB_UTIL_AUTOCOMP_H
#define LIB_UTIL_AUTOCOMP_H

#include "lib/util_opt.h"

void generate_autocomp(struct util_opt *opt_vec, char *tool_name);

#endif
