/*
 * Page related definitions and functions.
 *
 * Copyright IBM Corp. 2023
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef BOOT_PAGE_H
#define BOOT_PAGE_H
#include "lib/zt_common.h"

#define PAGE_SIZE		_AC(4096, UL)
#endif
