/*
 * zdev - Modify and display the persistent configuration of devices
 *
 * Copyright IBM Corp. 2022
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

#ifndef ZDEV_ID_H
#define ZDEV_ID_H

#define ZDEV_SITE_ID_FILE    "/run/zdev_id.env"

#endif /* ZDEV_ID_H */
