/*
 * dump2tar - tool to dump files and command output into a tar archive
 *
 * Global variables
 *
 * Copyright IBM Corp. 2016, 2017
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 *
 */

#include "global.h"

/* Global settings */
bool global_threaded;
bool global_debug;
bool global_verbose;
bool global_quiet;
bool global_timestamps;
